#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
Example of downloading map tiles and then rendering them as a map.
"""

import logging
logging.basicConfig(level=logging.DEBUG)

import asyncio
import geotiler

def log_error(tile):
    if tile.error:
        print('tile {} error: {}'.format(tile.url, tile.error))
    return tile

async def fetch(mm):
    tiles = geotiler.fetch_tiles(mm)
    # process tile in error and then render all the tiles
    tiles = (log_error(t) async for t in tiles)
    img = await geotiler.render_map_async(mm, tiles=tiles)
    return img


bbox = 11.78560, 46.48083, 11.79067, 46.48283
mm = geotiler.Map(extent=bbox, zoom=18)

loop = asyncio.get_event_loop()

img = loop.run_until_complete(fetch(mm))
img.save('ex-fetch-tiles.png', 'png')

# vim: sw=4:et:ai



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

import numpy as np
from functools import partial
from geotiler.map import Map, _find_top_left_tile, _tile_coords, _tile_offsets

import pytest
import unittest


approx = partial(pytest.approx, abs=1e-6)


class MapTestCase(unittest.TestCase):
    """
    Map unit tests.
    """
    def test_map_create_error_center_extent(self):
        """
        Test map instantiation error with center and extent
        """
        extent = (0, 0, 0, 0)
        center = (0, 0)
        map = self.assertRaises(ValueError, Map, extent=extent, center=center)


    def test_map_create_error_extent_size_zoom(self):
        """
        Test map instantiation error with extent, size and zoom
        """
        extent = (0, 0, 0, 0)
        size = (0, 0)
        zoom = 1
        map = self.assertRaises(
            ValueError, Map, extent=extent, size=size, zoom=zoom
        )


    def test_map_create_extent_size(self):
        """
        Test map instantiation with extent and size
        """
        extent = 11.785390377044687, 46.4799402452901, 11.790883541107167, 46.48372275323265
        size = 512, 512
        map = Map(extent=extent, size=size)

        self.assertEqual((512, 512), map.size)
        self.assertEqual(17, map.zoom)
        assert extent == approx(map.extent)
        assert (11.788137, 46.481832) == approx(map.center)

        self.assertEqual((69827, 46376), map.origin)
        self.assertEqual((-238, -194), map.offset)


    def test_map_create_extent_zoom(self):
        """
        Test map instantiation with extent and zoom
        """
        extent = 11.785604953765853, 46.48083418203029, 11.790668964386, 46.4828288639531
        zoom = 17
        map = Map(extent=extent, zoom=zoom)

        self.assertEqual((472, 270), map.size)
        self.assertEqual(17, map.zoom)
        assert extent == approx(map.extent)
        assert (11.788137, 46.481832) == approx(map.center)

        self.assertEqual((69827, 46376), map.origin)
        self.assertEqual((-238, -194), map.offset)


    def test_map_size_change_954_541(self):
        """
        Test changing map size (954x541)
        """
        extent = 11.785604953765853, 46.48083418203029, 11.790668964386, 46.4828288639531
        zoom = 17
        map = Map(extent=extent, zoom=zoom)

        assert map.size == (472, 270), map.size

        map.size = 945, 541

        self.assertEqual((945, 541), map.size)
        self.assertEqual(17, map.zoom) # zoom has not changed
        expected = 11.783067, 46.479833, 11.793206, 46.483829
        assert approx(expected) == map.extent

        self.assertEqual((69827, 46376), map.origin)
        self.assertEqual((-238, -194), map.offset)


    def test_map_size_change_1908x1082(self):
        """
        Test changing map size (1908x1082)
        """
        extent = 11.785604953765853, 46.48083418203029, 11.790668964386, 46.4828288639531
        zoom = 17
        map = Map(extent=extent, zoom=zoom)

        assert map.size == (472, 270), map.size

        map.size = 1908, 1082

        self.assertEqual((1908, 1082), map.size)
        self.assertEqual(17, map.zoom) # zoom has not changed
        expected = 11.777902, 46.4778346, 11.7983723, 46.4858281
        for v1, v2 in zip(expected, map.extent):
            self.assertAlmostEqual(v1, v2, 6)

        self.assertEqual((69827, 46376), map.origin)
        self.assertEqual((-238, -194), map.offset)


    def test_map_center_change(self):
        """
        Test changing map center
        """
        extent = 11.785604953765853, 46.48083418203029, 11.790668964386, 46.4828288639531
        zoom = 17
        map = Map(extent=extent, zoom=zoom)

        assert map.size == (472, 270), map.size

        delta = 4 * 1e-5
        map.center = 11.788136959075942 + delta, 46.481831532133 + delta

        self.assertEqual((472, 270), map.size)
        self.assertEqual(17, map.zoom)

        expected = 11.7856478, 46.480871, 11.790712, 46.482866
        for v1, v2 in zip(expected, map.extent):
            self.assertAlmostEqual(v1, v2, 6, map.extent)

        self.assertEqual((69827, 46376), map.origin)
        self.assertEqual((-242, -189), map.offset)


    def test_map_zoom_change(self):
        """
        Test changing map zoom
        """
        extent = 11.785604953765853, 46.48083418203029, 11.790668964386, 46.4828288639531
        zoom = 17
        map = Map(extent=extent, zoom=zoom)

        assert map.size == (472, 270), map.size
        assert map.zoom == 17, map.zoom

        map.zoom = 16

        self.assertEqual((472, 270), map.size)
        self.assertEqual(16, map.zoom)

        expected = 11.783072, 46.479836, 11.793200, 46.483826
        assert approx(expected) == map.extent

        self.assertEqual((34913, 23188), map.origin)
        self.assertEqual(16, map.zoom)
        self.assertEqual((-247, -97), map.offset)


    def test_map_extent_change(self):
        """
        Test changing map extent
        """
        extent = 11.785604953765853, 46.48083418203029, 11.790668964386, 46.4828288639531
        zoom = 17
        map = Map(extent=extent, zoom=zoom)

        assert map.size == (472, 270), map.size
        assert map.zoom == 17, map.zoom

        map.extent = 11.785590, 46.480830, 11.790680, 46.482840

        self.assertEqual((474, 272), map.size)
        self.assertEqual(17, map.zoom)

        self.assertEqual((69827, 46376), map.origin)
        self.assertEqual((-238, -194), map.offset)


    def test_map_top_left_offset_calculation(self):
        """
        Test calculation of offset of top-left map tile
        """
        center = 11.788137, 46.481832
        zoom = 17
        size = 3000, 3000
        map = Map(center=center, zoom=zoom, size=size)

        assert map.size == (3000, 3000), map.size
        assert map.zoom == 17, map.zoom
        assert map.origin == (69827, 46376), map.origin
        assert map.offset == (-238, -194), map.offset

        coord, offset = _find_top_left_tile(map)
        self.assertEqual(69822.000, coord[0])
        self.assertEqual(46370.000, coord[1])
        self.assertEqual((-18, -230), offset)


    def test_map_top_left_offset_calculation_12(self):
        """
        Test calculation of offset of top-left map tile (zoom 12)
        """
        center = 11.788137, 46.481832
        zoom = 12
        size = 2000, 2000
        map = Map(center=center, zoom=zoom, size=size)

        assert map.size == (2000, 2000), map.size
        assert map.zoom == 12, map.zoom
        assert map.origin == (2182, 1449), map.origin
        assert map.offset == (-31, -70), map.offset

        coord, offset = _find_top_left_tile(map)
        self.assertEqual(2178, coord[0], coord)
        self.assertEqual(1445, coord[1], coord)
        self.assertEqual((-55, -94), offset)


    def test_map_tile_coords(self):
        """
        Test calculation of coordinates of map tiles
        """
        center = 11.788137, 46.481832
        zoom = 17
        size = 300, 300
        map = Map(center=center, zoom=zoom, size=size)

        assert map.origin == (69827, 46376), map.origin
        assert map.offset == (-238, -194), map.offset

        coord, offset = _find_top_left_tile(map)
        coords = tuple(_tile_coords(map, coord, offset))

        c1, c2, c3, c4 = coords

        # first col
        self.assertEqual((69827, 46376), c1, coords)
        self.assertEqual((69827, 46377), c2, coords)
        # second col
        self.assertEqual((69828, 46376), c3, coords)
        self.assertEqual((69828, 46377), c4, coords)


    def test_map_tile_offsets(self):
        """
        Test calculation of image offsets of map tiles
        """
        center = 11.788137, 46.481832
        zoom = 17
        size = 300, 300
        map = Map(center=center, zoom=zoom, size=size)

        assert map.origin == (69827, 46376), map.origin
        assert map.offset == (-238, -194), map.offset

        coord, offset = _find_top_left_tile(map)
        offsets = tuple(_tile_offsets(map, offset))

        o1, o2, o3, o4 = offsets

        # first col
        self.assertEqual((-88, -44), o1, offsets)
        self.assertEqual((-88, -44 + 256), o2, offsets)
        # second col
        self.assertEqual((-88 + 256, -44), o3, offsets)
        self.assertEqual((-88 + 256, -44 + 256), o4, offsets)


    def test_rev_geocode(self):
        """
        Test map reverse geocode
        """
        location = -6.066, 53.386
        map = Map(center=location, zoom=15, size=(1000, 1000))
        pos = map.rev_geocode(map.center)
        self.assertEqual((500, 500), pos)


def test_map_create_center_zoom_size():
    """
    Test map instantiation with center, zoom and size
    """
    center = 11.788137, 46.481832
    zoom = 17
    size = 512, 512
    map = Map(center=center, zoom=zoom, size=size)
    assert (512, 512) == map.size
    assert 17 == map.zoom

    expected = 11.785390, 46.479940, 11.790883, 46.483722
    assert expected == approx(map.extent)
    assert (11.788137, 46.481832) == approx(map.center)

    assert (69827, 46376) == map.origin
    assert (-238, -194) == map.offset

def test_map_create_center_zoom_size_numpy():
    """
    Test map instantiation with center, zoom and size using NumPy arrays
    """
    center = np.array([11.788137, 46.481832])
    zoom = 17
    size = np.array([512, 512])
    map = Map(center=center, zoom=zoom, size=size)
    assert np.all([512, 512] == map.size)
    assert 17 == map.zoom

    expected = 11.785390, 46.479940, 11.790883, 46.483722
    assert expected == approx(map.extent)
    assert (11.788137, 46.481832) == approx(map.center)

    assert (69827, 46376) == map.origin
    assert (-238, -194) == map.offset

def test_map_zoom_change_same():
    """
    Test if extent and center of map holds after zoom reset to same value.
    """
    center = (11.788137, 46.481832)
    zoom = 15
    size = 512, 512
    map = Map(center=center, zoom=zoom, size=size)

    # test preconditions
    assert approx((11.788158, 46.481846)) == map.center
    assert approx((11.777172, 46.474280, 11.799144, 46.489410)) == map.extent

    # reset zoom to the same value
    map.zoom = 15

    # the center and map extent have not changed
    assert approx((11.788158, 46.481846)) == map.center
    assert approx((11.777172, 46.474280, 11.799144, 46.489410)) == map.extent

def test_map_create_error_size():
    """
    Test map instantiation error with size incorrect type
    """
    extent = 11.785390377044687, 46.4799402452901, 11.790883541107167, 46.48372275323265
    size = (512.0, 512.0)
    pytest.raises(TypeError, Map, extent=extent, size=size)

def test_map_error_set_size():
    """
    Test if error is raised when size has wrong type on size override
    """
    extent = 11.785390377044687, 46.4799402452901, 11.790883541107167, 46.48372275323265
    size = (512, 512)
    map = Map(extent=extent, size=size)
    with pytest.raises(TypeError):
        map.size = (512.0, 512.0)

# vim: sw=4:et:ai



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
Caching downloader unit tests.
"""

import asyncio
from functools import partial

from geotiler.map import Tile
from geotiler.cache import caching_downloader, redis_downloader

from unittest import mock

def test_redis_downloader_and_cache():
    """
    Test Redis downloader and cache functions.
    """
    async def images(tiles, num_workers):
        tile = partial(Tile, error=None, offset=None)
        for t in tiles:
            yield tile(url=t.url, img='img')

    async def as_list(tiles):
        return [t async for t in tiles]

    client = mock.MagicMock()
    client.get.side_effect = ['c-img1', None, 'c-img3']
    downloader = redis_downloader(client, downloader=images, timeout=10)
    assert caching_downloader == downloader.func

    urls = ['url1', 'url2', 'url3']
    tiles = [Tile(url, None, None, None) for url in urls]

    loop = asyncio.get_event_loop()
    tiles = downloader(tiles, 2)
    result = loop.run_until_complete(as_list(tiles))

    args = [v[0][0] for v in client.get.call_args_list]
    assert ['url1', 'url2', 'url3'] == args

    args = sorted(v[0] for v in client.setex.call_args_list)
    assert 3 == len(args)
    assert ('url1', 10, 'c-img1') == args[0]
    assert ('url2', 10, 'img') == args[1]
    assert ('url3', 10, 'c-img3') == args[2]


# vim: sw=4:et:ai



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
Tests for rendering map image using map tile data.
"""

import asyncio
import io
import PIL.Image  # type: ignore

from geotiler.map import Tile
import geotiler.tile.img as tile_img

from unittest import mock


def _run_render_image(map, tiles):
    """
    Run coroutine rendering map image using the map tiles.

    :param map: Map object.
    :param tiles: Asynchronous generator of tiles.
    """
    loop = asyncio.get_event_loop()
    task = tile_img.render_image(map, tiles)
    image = loop.run_until_complete(task)
    return image

async def _tile_generator(offsets, data):
    """
    Create asynchronous generator of map tiles.
    """
    for o, i in zip(offsets, data):
        yield Tile(None, o, i, None)

def test_render_error_tile():
    """
    Test rendering of error tile.
    """
    tile_img._error_image.cache_clear()
    img = tile_img._error_image(10, 10)
    assert (10, 10) == img.size

def test_tile_image_png():
    """
    Test converting PNG data into PIL image object.
    """
    tile = PIL.Image.new('RGBA', (10, 11))
    f = io.BytesIO()
    tile.save(f, format='png')

    img = tile_img._tile_image(f.getbuffer())
    assert (10, 11) == img.size

def test_tile_image_jpg():
    """
    Test converting JPEG data into PIL image object.
    """
    tile = PIL.Image.new('RGB', (12, 10))
    f = io.BytesIO()
    tile.save(f, format='jpeg')

    img = tile_img._tile_image(f.getbuffer())
    assert (12, 10) == img.size

def test_render_image():
    """
    Test rendering map image.
    """
    tile = PIL.Image.new('RGBA', (10, 10))
    map = mock.MagicMock()
    map.size = [30, 20]
    map.provider.tile_width = 10
    map.provider.tile_height = 10

    with mock.patch('geotiler.tile.img._tile_image') as tf, \
            mock.patch.object(PIL.Image, 'new') as img_new:

        tf.return_value = tile
        data = (tile, tile, tile, tile)
        offsets = ((0, 0), (10, 0), (20, 0), (0, 10))
        tiles = _tile_generator(offsets, data)
        image = _run_render_image(map, tiles)
        img_new.assert_called_once_with('RGBA', (30, 20))
        assert 4 == tf.call_count
        tf.assert_called_with(tile)

def test_render_image_error():
    """
    Test rendering map image with error tile.
    """
    tile = PIL.Image.new('RGBA', (10, 10))
    map = mock.MagicMock()
    map.size = 30, 20
    map.provider.tile_width = 10
    map.provider.tile_height = 10

    with mock.patch('geotiler.tile.img._tile_image') as tf:
        tf.return_value = tile
        data = (tile, tile, None, tile, None, tile)
        offsets = ((0, 0), (10, 0), (20, 0), (0, 10), (10, 10), (20, 10))
        tiles = _tile_generator(offsets, data)
        image = _run_render_image(map, tiles)
        assert 4 == tf.call_count
        tf.assert_called_with(tile)

# vim: sw=4:et:ai



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
Map tile downloading unit tests.
"""

import asyncio
import aiohttp
from contextlib import contextmanager

import geotiler.tile.io
from geotiler.map import Tile
from geotiler.tile.io import fetch_tile, fetch_tiles

import pytest
from unittest import mock

@contextmanager
def mock_url_open(session, data, error_msg=None):
    """
    Mock opening of an URL.

    The following call needs to be mocked::

        ClientSession.get().__aenter__().read()

    If error message set, then it is raised as `aiohttp.ClientError`
    exception.

    :param session: Mock of `aiohttp` client session.
    :param data: Data to be assigned to HTTP response.
    :param error_msg: Error encountered when an exception is raised.
    """
    mock_get = session.get.return_value
    mock_ctx = mock_get.__aenter__.return_value = mock.MagicMock()

    if error_msg:
        params = {'side_effect': aiohttp.ClientError(error_msg)}
    else:
        params = {'return_value': data}

    mock_ctx.read = mock.AsyncMock(**params)
    yield session

@pytest.mark.asyncio
@mock.patch('aiohttp.ClientSession')
async def test_fetch_tile(session):
    """
    Test fetching a map tile.
    """
    tile = Tile('http://a.b.c', None, None, None)

    with mock_url_open(session, 'image'):
        result = await fetch_tile(session, tile)

        assert result is not tile  # copy of tile is returned
        assert tile.url == result.url
        assert 'image' == result.img
        assert result.error is None

@pytest.mark.asyncio
@mock.patch('aiohttp.ClientSession')
async def test_fetch_tile_error(session):
    """
    Test fetching a map tile with an error.
    """
    # assign something to `img` and `error` to see if they are overriden
    # properly in case of an error
    tile = Tile('http://a.b.c', 'a', 'b', 'c')
    with mock_url_open(session, 'image', error_msg='some error'):
        tile = await fetch_tile(session, tile)
        assert tile.img is None

        error = 'Unable to download http://a.b.c (error: some error)'
        assert error == str(tile.error)

@pytest.mark.asyncio
@mock.patch('aiohttp.ClientSession')
async def test_fetch_tiles(session):
    """
    Test fetching map tiles.
    """
    tiles = [
        Tile('http://a.b.c/1', 'o', 'image', None),
        Tile('http://a.b.c/2', 'o', 'image', None),
        Tile('http://a.b.c/3', 'o', None, 'error'),
        Tile('http://a.b.c/4', 'o', 'image', None),
    ]
    tasks = [asyncio.Future() for t in tiles]
    for task, tile in zip(tasks, tiles):
        task.set_result(tile)

    ctx_ac = mock.patch.object(asyncio, 'as_completed')
    ctx_ft = mock.patch.object(geotiler.tile.io, 'fetch_tile')
    with ctx_ac as mock_as_completed, ctx_ft:
        mock_as_completed.return_value = tasks
        tiles = [t async for t in fetch_tiles(tiles, 2)]

        img = [tile.img for tile in tiles]
        assert ['image', 'image', None, 'image'] == img, tiles

        error = [tile.error for tile in tiles]
        assert [None, None, 'error', None] == error, tiles

# vim: sw=4:et:ai



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
Read data from GPS and show map centered at the position.

Use keys like '+', '-', 'n', 'p', 'r' to zoom in, zoom out, change map
provider and refresh map.

Requirements:

- geotiler
- quamash
- PyQt 5
"""

import asyncio
import functools
import json
import logging
import redis
import sys
from collections import deque

from PIL.ImageQt import ImageQt

from PyQt5 import QtCore
from PyQt5.QtWidgets import QApplication, QGraphicsView, QGraphicsScene, \
    QGraphicsPixmapItem, QGraphicsEllipseItem
from PyQt5.QtGui import QImage, QPixmap, QPainter, QPen, QColor
from quamash import QEventLoop

import geotiler
from geotiler.cache import redis_downloader

logging.getLogger('geotiler').setLevel(logging.DEBUG)
logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def scroll_map(widget, pos):
    """
    Center map image on current position.

    :param map: Geotiler map object.
    :param pos: Current position (longitude, latitude).
    """
    map = widget.map
    widget.position = pos

    w1, h1 = widget.width() / 2, widget.height() / 2
    w2, h2 = map.size
    x, y = map.rev_geocode(widget.position)

    within = 0 <= x - w1 and x + w1 < w2 \
         and 0 <= y - h1 and y + h1 < h2
    if not within:
        map.center = pos
        widget.refresh_map.set()
        x, y = w2 / 2, h2 / 2

    widget.centerOn(x, y)


async def read_gps(queue):
    """
    Read location from gpsd daemon and put it into positions queue.

    :param queue: Queue of GPS positions.
    """
    reader, writer = await asyncio.open_connection(port=2947)
    writer.write('?WATCH={"enable":true,"json":true}\n'.encode())
    while True:
        line = await reader.readline()
        data = json.loads(line.decode())
        if 'lon' in data:
            await queue.put((data['lon'], data['lat']))


async def refresh_map(widget):
    """
    Refresh map when map widget refresh event is set.

    This is asyncio coroutine.

    :param widget: Map widget.
    """
    event = widget.refresh_map
    map = widget.map

    # use redis to cache map tiles
    client = redis.Redis('localhost')
    downloader = redis_downloader(client)
    render_map = functools.partial(
        geotiler.render_map_async, downloader=downloader
    )
    fetch_tiles = functools.partial(
        geotiler.fetch_tiles, downloader=downloader
    )

    pixmap = QPixmap(*map.size)

    while True:
        await event.wait()
        event.clear()

        logger.debug('fetching map image...')

        img = await render_map(map)
        pixmap.convertFromImage(ImageQt(img))

        # TODO: use `fetch_tiles` to update map as tiles arrive, but try to
        # avoid `setPixmap` within the loop.
        # tiles = fetch_tiles(map)
        # async for tile in tiles:
        #     painter = QPainter(pixmap)
        #     img = QImage()
        #     img.loadFromData(tile.img)
        #     painter.drawImage(*tile.offset, img)
        #     painter.end()
        #     widget.map_layer.setPixmap(pixmap)

        scroll_map(widget, map.center)
        widget.map_layer.setPixmap(pixmap)

        logger.debug('got map image')


async def locate(widget, queue):
    """
    Read position from the queue and update map position.

    This is asyncio coroutine.

    :param widget: Map widget.
    :param queue: Queue of GPS positions.
    """
    while True:
        pos = await queue.get()
        scroll_map(widget, pos)


class MapWindow(QGraphicsView):
    """
    Qt widget based on QGraphicsView to display map image and control map
    with keyboard.

    :var map: Map object.
    :var position: Current position on the map.
    """
    def __init__(self, map, *args):
        super().__init__(*args)
        self.map = map
        self.position = map.center

        self.providers = deque([
            'osm', 'stamen-terrain', 'stamen-toner-lite', 'stamen-toner',
            'stamen-watercolor', 'ms-aerial', 'ms-hybrid', 'ms-road',
            'bluemarble',
        ])

        self.refresh_map = asyncio.Event()

        scene = QGraphicsScene()
        self.scene = scene
        self.setScene(scene)
        self.map_layer = QGraphicsPixmapItem()
        scene.addItem(self.map_layer)

        self.circle = QGraphicsEllipseItem(0, 0, 20, 20)
        pen = QPen(QColor(255, 0, 0, 128))
        pen.setWidth(2)
        self.circle.setPen(pen)
        scene.addItem(self.circle)

        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)


    def centerOn(self, x, y):
        """
        Center on (x, y) coordinates, which reflect current map position
        and draw current position marker.
        """
        super().centerOn(x, y)
        self.circle.setRect(x - 10, y - 10, 20, 20)


    def keyPressEvent(self, event):
        """
        Handle key press events

        - '-' to zoom out
        - '+'/'=' to zoom in
        - 'n' to use next map provider
        - 'p' to use previous map provider
        - 'r' to refresh map
        - 'q' to quit

        """
        key = event.key()
        if key == QtCore.Qt.Key_Minus:
            self.map.zoom -= 1
            self.refresh_map.set()
        elif key == QtCore.Qt.Key_Plus or key == QtCore.Qt.Key_Equal:
            self.map.zoom += 1
            self.refresh_map.set()
        elif key == QtCore.Qt.Key_Q:
            loop.stop()
        elif key == QtCore.Qt.Key_N:
            self.providers.rotate(-1)
            p = self.providers[0]
            self.map.provider = geotiler.find_provider(p)
            self.refresh_map.set()
        elif key == QtCore.Qt.Key_P:
            self.providers.rotate(1)
            p = self.providers[0]
            self.map.provider = geotiler.find_provider(p)
            self.refresh_map.set()
        elif key == QtCore.Qt.Key_R:
            self.refresh_map.set()


    def resizeEvent(self, event):
        """
        On resize event redraw the map.
        """
        scroll_map(self, self.position)



# bind Qt application with Python asyncio loop
app = QApplication(sys.argv)
loop = QEventLoop(app)
asyncio.set_event_loop(loop)

g = app.desktop().screenGeometry()
size = g.width() + 256 * 2, g.height() + 256 * 2

pos = (0, 0)
if len(sys.argv) == 3:
    pos = float(sys.argv[1]), float(sys.argv[2])

mm = geotiler.Map(size=size, center=pos, zoom=18)

window = MapWindow(mm)
window.show()

queue = asyncio.Queue()
t1 = read_gps(queue)
t2 = locate(window, queue)
t3 = refresh_map(window)
task = asyncio.gather(t1, t2, t3)

window.refresh_map.set()
loop.run_until_complete(task)

# vim: sw=4:et:ai



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

from geotiler.provider import MapProvider, base_dir, obfuscate

from unittest import mock


def test_provider_init_default():
    """
    Test initialization of map provider with default configuration values.
    """
    data = {
        'name': 'OpenStreetMap',
        'attribution': '© OpenStreetMap contributors\nhttp://www.openstreetmap.org/copyright',
        'url': 'http://{subdomain}.tile.openstreetmap.org/{z}/{x}/{y}.{ext}',
    }
    provider = MapProvider(data)

    assert 'OpenStreetMap' == provider.name
    expected = '© OpenStreetMap contributors' \
        + '\nhttp://www.openstreetmap.org/copyright'
    assert expected == provider.attribution
    expected = 'http://{subdomain}.tile.openstreetmap.org/{z}/{x}/{y}.{ext}'
    assert expected == provider.url
    assert () == provider.subdomains
    assert 'png' == provider.extension
    assert 1 == provider.limit
    assert provider.api_key_ref is None

def test_provider_init_default_override():
    """
    Test initialization of map provider when overriding default
    configuration values.
    """
    data = {
        'name': 'OpenStreetMap',
        'attribution': '© OpenStreetMap contributors\nhttp://www.openstreetmap.org/copyright',
        'url': 'http://{subdomain}.tile.openstreetmap.org/{z}/{x}/{y}.{ext}',
        'subdomains': ('a', 'b', 'c'),
        'extension': 'jpg',
        'limit': 2,
        'api-key-ref': 'a-b-c',
        'tile-width': 513,
        'tile-height': 514,
    }
    provider = MapProvider(data)

    assert 'OpenStreetMap' == provider.name
    expected = '© OpenStreetMap contributors' \
        + '\nhttp://www.openstreetmap.org/copyright'
    assert expected == provider.attribution
    expected = 'http://{subdomain}.tile.openstreetmap.org/{z}/{x}/{y}.{ext}'
    assert expected == provider.url
    assert ('a', 'b', 'c') == provider.subdomains
    assert 'jpg' == provider.extension
    assert 2 == provider.limit
    assert 'a-b-c' == provider.api_key_ref
    assert 513 == provider.tile_width
    assert 514 == provider.tile_height

def test_provider_tile_url():
    """
    Test map provider tile url formatting.
    """
    data = {
        'name': 'OpenStreetMap',
        'attribution': '© OpenStreetMap contributors\nhttp://www.openstreetmap.org/copyright',
        'subdomains': ['a'],
        'url': 'http://{subdomain}.tile.openstreetmap.org/{z}/{x}/{y}.{ext}',
    }
    provider = MapProvider(data)
    url = provider.tile_url((1, 2), 15)
    assert 'http://a.tile.openstreetmap.org/15/1/2.png' == url

def test_provider_tile_url_api_key():
    """
    Test map provider tile url formatting with API key.
    """
    data = {
        'name': 'OpenStreetMap',
        'attribution': '© OpenStreetMap contributors\nhttp://www.openstreetmap.org/copyright',
        'subdomains': ['a'],
        'url': 'http://{subdomain}.tile.openstreetmap.org/{z}/{x}/{y}.{ext}?apikey={api_key}',
        'api-key-ref': 'a-key-ref',
    }
    provider = MapProvider(data, api_key='a-key-ref')
    url = provider.tile_url((1, 2), 15)
    assert 'http://a.tile.openstreetmap.org/15/1/2.png?apikey=a-key-ref' == url

def test_base_dir():
    """
    Test base dir retrieval.
    """
    fn = base_dir()
    assert fn.endswith('/geotiler/source')

def test_obfuscate_url():
    """
    Test obfuscation of API key in a tile URL.
    """
    # no api key, no change
    url = obfuscate('http://a.tile.openstreetmap.org/15/1/2.png')
    assert 'http://a.tile.openstreetmap.org/15/1/2.png' == url

    url = obfuscate('http://a.tile.openstreetmap.org/15/1/2.png?apikey=a-key-ref')
    assert 'http://a.tile.openstreetmap.org/15/1/2.png?apikey=<apikey>' == url

    url = obfuscate('http://a.tile.openstreetmap.org/15/1/2.png?apikey=a-key-ref&a=1&b=2')
    assert 'http://a.tile.openstreetmap.org/15/1/2.png?apikey=<apikey>&a=1&b=2' == url

    url = obfuscate('http://a.tile.openstreetmap.org/15/1/2.png?api-key=a-key-ref')
    assert 'http://a.tile.openstreetmap.org/15/1/2.png?api-key=<apikey>' == url

    url = obfuscate('http://a.tile.openstreetmap.org/15/1/2.png?api-key=a-key-ref&a=1&b=2')
    assert 'http://a.tile.openstreetmap.org/15/1/2.png?api-key=<apikey>&a=1&b=2' == url

# vim:et sts=4 sw=4:



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

from geotiler.geo import Transformation, WebMercator, zoom_to
from functools import partial

import pytest

approx = partial(pytest.approx, abs=1e-3)

@pytest.fixture
def web_mercator():
    wm = WebMercator(10)
    wm.transformation = Transformation(1, 0, 0, 0, 1, 0)
    return wm


def test_transformation_identity():
    """
    Test identity transformation.
    """
    t = Transformation(1, 0, 0, 0, 1, 0)
    p = 1, 1

    pt = t.transform(p)
    assert 1.0 == pt[0]
    assert 1.0 == pt[1]

    ptt = t.untransform(pt)
    assert 1.0 == pt[0]
    assert 1.0 == pt[1]

def test_transformation_0_1_0():
    """
    Test transformation for (0, 1, 0, ...).
    """
    t = Transformation(0, 1, 0, 1, 0, 0)
    p = 0, 1

    pt = t.transform(p)
    assert 1.0 == pt[0]
    assert 0.0 == pt[1]

    ptt = t.untransform(pt)
    assert 0.0 == ptt[0]
    assert 1.0 == ptt[1]

def test_transformation_1_0_1():
    """
    Test transformation for (1, 0, 1, ...).
    """
    t = Transformation(1, 0, 1, 0, 1, 1)
    p = 0, 0

    pt = t.transform(p)
    assert 1.0 == pt[0]
    assert 1.0 == pt[1]

    ptt = t.untransform(pt)
    assert 0.0 == ptt[0]
    assert 0.0 == ptt[1]

def test_web_mercator_rev_geocode_at_zero(web_mercator):
    """
    Test Web Mercator projection reverse geocode at (0, 0) position.
    """
    coord = web_mercator.rev_geocode((0, 0))
    assert 0.0 == approx(coord[0])
    assert 0.0 == approx(coord[1])

def test_web_mercator_geocode_at_zero(web_mercator):
    """
    Test Web Mercator projection geocode at (0, 0) position.
    """
    pt = web_mercator.geocode((0, 0), 10)
    assert 0.0 == approx(pt[0])
    assert 0.0 == approx(pt[1])

def test_web_mercator_rev_geocode(web_mercator):
    """
    Test Web Mercator projection reverse geocode at a position.
    """
    coord = web_mercator.rev_geocode((-122, 37))
    assert -2.129 == approx(coord[0])
    assert 0.696 == approx(coord[1])

def test_web_mercator_geocode(web_mercator):
    """
    Test Web Mercator projection geocode at a position.
    """
    pt = web_mercator.geocode((-2.129, 0.696), 10.000)
    assert -121.983 == approx(pt[0])
    assert 37.001 == approx(pt[1])

def test_zoom():
    """
    Test zooming tile coordinates.
    """
    coord = zoom_to((1, 0), 2, 3)
    assert (2, 0) == coord

    coord = zoom_to((1, 0), 2, 1)
    assert (0.5, 0) == coord

# vim: sw=4:et:ai



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
GPS example.

Requires running `gpsd` daemon.
"""

import asyncio
import functools
import logging
import json
import redis

logging.basicConfig(level=logging.DEBUG)

import geotiler
from geotiler.cache import redis_downloader

client = redis.Redis('localhost')
downloader = redis_downloader(client)
render_map_async = functools.partial(
    geotiler.render_map_async, downloader=downloader
)

async def read_gps(queue):
    """
    Read location data from `gpsd` daemon.
    """
    reader, writer = await asyncio.open_connection(port=2947)
    writer.write(b'?WATCH={"enable":true,"json":true}\n')
    while True:
        line = await reader.readline()
        data = json.loads(line.decode())
        if 'lon' in data:
            await queue.put((data['lon'], data['lat']))
        
async def show_map(queue, map):
    """
    Save map centered at location to a file.
    """
    while True:
        pos = yield from queue.get()

        map.center = pos
        img = await render_map_async(map)
        img.save('ex-async-gps.png', 'png')


size = 800, 800
start = 0, 0
mm = geotiler.Map(size=size, center=start, zoom=16)

queue = asyncio.Queue(1)        # queue holding current position from gpsd

# run location and map rendering tasks concurrently
t1 = show_map(queue, mm)
t2 = read_gps(queue)
task = asyncio.gather(t1, t2)
loop = asyncio.get_event_loop()
loop.run_until_complete(task)

# vim: sw=4:et:ai



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

import geotiler
from functools import partial
from geotiler.cache import caching_downloader
from geotiler.tile.io import fetch_tiles
import shelve
import time


#
# Make it default GeoTiler cache? We need to
#
# - create shelve file per map provider in ~/.geotiler/ directory
# - figure out key timestamp issue
# - unit tests
#
class Cache:
    """
    Cache based on Python's `shelve` module to keep map tiles.

    WARNING: There is nothing that prevents cache from growing very large
    and also nothing which determines if the data is recent (however the
    time at the moment something is placed in the database is saved).
    """
    def __init__(self, filename):
        """
        Opens an existing shelve and create new one if it does not exist.

        :param filename: Filename for the shelve.
        """
        self.cache = shelve.open(filename, writeback=True)


    def get(self, key):
        """
        Get map tile from cache.

        Null is returned if key does not exist.

        :param key: Key used to retrieve data from the cache.
        """
        try:
            data = self.cache[key][0]
        except:
            data = None
        return data


    def set(self, key, data):
        """
        Stores map tile in cache.

        The data is stored as (tile data, time) tuple.

        :param key: Key used to identify the data.
        :param data: Map tile data.
        """
        if key not in self.cache:
            self.cache[key] = (data, time.time())
        return


    def close(self):
        """
        Close the shelve used to cache the data.
        """
        self.cache.close()


# Create a cache instance
cache = Cache("test")

# Create a new downloader using the shelve cache
downloader = partial(caching_downloader, cache.get, cache.set, fetch_tiles)

# Set render_map to use new downloader function
render_map = partial(geotiler.render_map, downloader=downloader)

mm = geotiler.Map(center=(0.0, 51.47879), zoom=18, size=(1900, 1000))

# Render map for first time and save current time
t1 = time.time()
img = render_map(mm)
t2 = time.time()

# Render map for the second time to demonstrate the speed of the cache
img = render_map(mm)
t3 = time.time()

# Close the cache
cache.close()

print('time for first render: {:6.5f} seconds'.format(t2 - t1))
print('time for second render: {:6.5f} seconds'.format(t3 - t2))

# vim: sw=4:et:ai



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
Matplotlib example.
"""

import matplotlib.pyplot as plt

import logging
logging.basicConfig(level=logging.DEBUG)

import geotiler

bbox = 11.78560, 46.48083, 11.79067, 46.48283

fig = plt.figure(figsize=(10, 10))
ax = plt.subplot(111)

#
# download background map using OpenStreetMap
#
mm = geotiler.Map(extent=bbox, zoom=18)

img = geotiler.render_map(mm)
ax.imshow(img)

#
# plot custom points
#
x0, y0 = 11.78816, 46.48114 # http://www.openstreetmap.org/search?query=46.48114%2C11.78816
x1, y1 = 11.78771, 46.48165 # http://www.openstreetmap.org/search?query=46.48165%2C11.78771
points = ((x0, y0), (x1, y1))
x, y = zip(*(mm.rev_geocode(p) for p in points))
ax.scatter(x, y, c='red', edgecolor='none', s=10, alpha=0.9)

plt.savefig('ex-matplotlib.pdf', bbox_inches='tight')
plt.close()

# vim: sw=4:et:ai



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

import math

class Projection:
    """
    Map projection.

    :var zoom: Map zoom.
    :var transformation: Coordinates transformation.
    :var max_extent: Maximum extent of transformed map.
    """
    def __init__(self, zoom):
        self.zoom = zoom
        self.transformation = Transformation(1, 0, 0, 0, 1, 0)
        self.max_extent = (-180, -90, 180, 90)

    def project(self, point):
        raise NotImplementedError()

    def unproject(self, point):
        raise NotImplementedError()

    def rev_geocode(self, location):
        """
        Reverse geocode location as tile coordinates at projection's zoom.

        The method returns tile coordinates (x, y) for given location and
        projection's zoom level.

        :param location: Location to reverse geocode.
        """
        x, y = location
        point = math.radians(x), math.radians(y)
        point = self.project(point)
        return self.transformation.transform(point)

    def geocode(self, tile_coord, zoom):
        """
        Geocode tile coordinates and zoom level information.

        The method returns (longitude, latitude) pair of the tile
        coordinates at their zoom level.

        :param tile_coord: Tile coordinates.
        :param zoom: Zoom of the tile coordinates.
        """
        pt = zoom_to(tile_coord, zoom, self.zoom)
        pt = self.transformation.untransform(pt)
        x, y = self.unproject(pt)
        return math.degrees(x), math.degrees(y)

class WebMercator(Projection):
    """
    Web Mercator projection.

    .. seealso:: http://epsg.io/3857
    .. seealso:: https://en.m.wikipedia.org/wiki/Web_Mercator_projection
    .. seealso:: Geomatics Guidance Note 7, part 2 Coordinate Conversions
                 & Transformations including Formulas, page 42
    .. seealso:: Implementation Practice Web Mercator Map Projection,
                 NGA.SIG.0011_1.0.0_WEBMERC, 2014-02-18, Version 1.0.0,
                 National Geospatial-Intelligence Agency (NGA)
    """
    def __init__(self, zoom):
        super().__init__(zoom)

        # the spherical mercator world tile covers (-π, -π) to (π, π)
        t = to_transformation(
            -math.pi, math.pi,
            0, 0,
            math.pi, math.pi,
            1, 0,
            -math.pi, -math.pi,
            0, 1,
        )
        self.transformation = t

        max_lat = 85.051129
        self.max_extent = (-180, -max_lat, 180, max_lat)

    def project(self, point):
        x, y = point
        return x, math.log(math.tan(0.25 * math.pi + 0.5 * y))

    def unproject(self, point):
        x, y = point
        return x, 2 * math.atan(math.pow(math.e, y)) - 0.5 * math.pi

class Transformation:
    def __init__(self, ax, bx, cx, ay, by, cy):
        self.ax = ax
        self.bx = bx
        self.cx = cx
        self.ay = ay
        self.by = by
        self.cy = cy

    def transform(self, point):
        x, y = point
        return (
            self.ax * x + self.bx * y + self.cx,
            self.ay * x + self.by * y + self.cy
        )

    def untransform(self, point):
        x, y = point
        return (
            (x * self.by - y * self.bx - self.cx * self.by + self.cy * self.bx)
            / (self.ax * self.by - self.ay * self.bx),
            (x * self.ay - y * self.ax - self.cx * self.ay + self.cy * self.ax)
            / (self.bx * self.ay - self.by * self.ax)
        )

def to_transformation(
        a1x, a1y,
        a2x, a2y,
        b1x, b1y,
        b2x, b2y,
        c1x, c1y,
        c2x, c2y
    ):
    """
    Generate a transform based on three pairs of points.

    The pairs of points are

    - a1 -> a2
    - b1 -> b2
    - c1 -> c2
    """
    ax, bx, cx = solve_linear(a1x, a1y, a2x, b1x, b1y, b2x, c1x, c1y, c2x)
    ay, by, cy = solve_linear(a1x, a1y, a2y, b1x, b1y, b2y, c1x, c1y, c2y)
    return Transformation(ax, bx, cx, ay, by, cy)

def solve_linear(r1, s1, t1, r2, s2, t2, r3, s3, t3):
    """
    Solve a system of linear equations.

    The equations are::

          t1 = (a * r1) + (b + s1) + c
          t2 = (a * r2) + (b + s2) + c
          t3 = (a * r3) + (b + s3) + c

    The `[rts][123]` are known values and `a`, `b` and `c` are the unknowns
    to be solved.

    Return values of `a`, `b` and `c`.
    """
    a = ((t2 - t3) * (s1 - s2) - (t1 - t2) * (s2 - s3)) \
      / ((r2 - r3) * (s1 - s2) - (r1 - r2) * (s2 - s3))

    b = ((t2 - t3) * (r1 - r2) - (t1 - t2) * (r2 - r3)) \
      / ((s2 - s3) * (r1 - r2) - (s1 - s2) * (r2 - r3))

    c = t1 - r1 * a - s1 * b

    return a, b, c

def zoom_to(tile_coord, zoom, target):
    """
    Zoom tile coordinates from current zoom to target zoom.

    :param tile_coord: Tile coordinates.
    :param zoom: Current zoom.
    :param target: Target zoom.
    """
    col, row = tile_coord
    d_zoom = target - zoom
    return col * math.pow(2, d_zoom), row * math.pow(2, d_zoom)

# vim:et sts=4 sw=4:



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
Caching strategies for GeoTiler.
"""

import logging
from functools import partial
from cytoolz.itertoolz import groupby, partition_all  # type: ignore

from .util import log_tiles
from geotiler.tile.io import fetch_tiles

logger = logging.getLogger(__name__)

def log_tile_cache_hit(tile):
    if tile.img:
        logger.debug('cache hit for: {}'.format(tile.url))
    return tile

def fetch_from_cache(get, tiles):
    tiles = (t._replace(img=get(t.url)) for t in tiles)
    if __debug__:
        tiles = log_tiles(log_tile_cache_hit, tiles)
    return tiles

async def caching_downloader(get, set, downloader, tiles, num_workers, **kw):
    """
    Download tiles from cache and missing tiles with the downloader.

    Asynchronous generator of map tiles is returned.

    The code flow is

    - caching downloader gets tile data from cache using URLs
    - the original downloader is used to download missing tile data
    - cache is updated with all existing tile data

    The cache getter function (`get` parameter) should return `None` if
    tile data is not in cache for given URL.

    A collection of tiles is returned.

    :param get: Function to get a tile data from cache.
    :param set: Function to put a tile data in cache.
    :param downloader: Original tiles downloader (asyncio coroutine).
    :param tiles: Collection tiles to fetch.
    :param num_workers: Number of workers used to connect to a map provider
        service.
    :param kw: Parameters passed to downloader coroutine.
    """
    tiles = fetch_from_cache(get, tiles)
    groups = partition_all(10, tiles)
    for tg in groups:
        missing = groupby(lambda t: t.img is None, tg)
        for t in missing.get(False, []):
            # reset cache for new and old tiles
            set(t.url, t.img)
            yield t

        result = downloader(missing.get(True, []), num_workers, **kw)
        async for t in result:
            # reset cache for new and old tiles
            set(t.url, t.img)
            yield t

def redis_downloader(client, downloader=None, timeout=3600 * 24 * 7):
    """
    Create downloader using Redis as cache for map tiles.

    :param client: Redis client object.
    :param downloader: Map tiles downloader, use `None` for default downloader.
    :param timeout: Map tile data expiry timeout, default 1 week.
    """
    if downloader is None:
        downloader = fetch_tiles
    set = lambda key, value: client.setex(key, timeout, value)
    return partial(caching_downloader, client.get, set, downloader)


# vim: sw=4:et:ai



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
GeoTiler map functionality.
"""

import asyncio
import itertools
import math
import numbers
import logging
import typing as tp
from collections import namedtuple

from .provider import DEFAULT_PROVIDER, find_provider, MapProvider
from .geo import zoom_to
from .tile.io import fetch_tiles as _fetch_tiles
from .tile.img import render_image
from .util import div_ceil

logger = logging.getLogger(__name__)

MAX_ZOOM = 25

Tile = namedtuple('Tile', ['url', 'offset', 'img', 'error'])
Tile.__doc__ = """
Map tile.

If a tile download succeeded, then `Tile.img` contain map tile image data.
Otherwise, it is set to null and `Tile.error` is set to tile download
error.

:var url: Tile URL.
:var offset: Tile offest in a map image.
:var img: Tile image data.
:var error: Tile error information.
"""

class Map:
    """
    Map created from tiles and to be drawn as an image.

    The extent, zoom, center and image size can be changed at any time with
    appropriate properties.

    :var provider: Map tiles provider identificator (default `osm`) or
        object.
    :var _zoom: Map zoom attribute accessed via `zoom` property.
    :var _size: Map image size accessed via `size` property.
    :var origin: Tile coordinates at map zoom level of base tile.
    :var offset: Position of base tile relative to map center.
    """
    def __init__(
        self, extent=None, center=None, zoom=None, size=None,
        provider: tp.Union[str, MapProvider]=DEFAULT_PROVIDER
    ) -> None:
        """
        Create map.

        One of the following parameters combination is required

        - center, zoom and size
        - extent and size
        - extent and zoom

        If none of above parameters combination is provided, then `ValueError`
        exception is raised.

        :param extent: Map geographical extent.
        :param center: Map geographical center.
        :param zoom: Map zoom.
        :param size: Map image size.
        :param provider: Map tiles provider.
        """
        super().__init__()

        self._check_size(size)

        if isinstance(provider, MapProvider):
            self.provider = provider
        else:
            self.provider = find_provider(provider)

        self.origin = None
        self.offset = None

        self._zoom = zoom
        self._size = size

        if center is not None and extent is not None:
            raise ValueError(
                'Bad map coverage, center and extent can\'t both be set'
            )
        elif extent is not None and size is not None and zoom is not None:
            raise ValueError(
                'Bad map coverage, size and zoom can\'t be set together' \
                ' with extent'
            )
        elif center is not None and zoom is not None and size is not None:
            self._change_center_zoom(center, zoom)
        elif extent is not None and size is not None:
            self._change_extent_and_zoom(extent)
        elif extent is not None and zoom is not None:
            self.extent = extent
        else:
            raise ValueError(
                'Unknown combination of extent, center, zoom and size' \
                ' parameters'
            )

        assert self._zoom is not None
        assert self._size is not None
        assert self.origin is not None
        assert self.offset is not None

        if __debug__:
            logger.debug(
                'extent={map.extent}, center={map.center}, zoom={map.zoom},'
                ' size={map.size}, origin={map.origin}, offset={map.offset}'
                .format(map=self)
            )


    @property
    def extent(self):
        """
        Calculate map geographical extent.

        It is a tuple of 2 coordinates (longitude and latitude)

        - lower-bottom corner of the map
        - right-top corner of the map

        Setting map extent changes map image size.
        """
        w, h = self._size
        p1 = self.geocode((0, h))
        p2 = self.geocode((w, 0))
        return p1[0], p1[1], p2[0], p2[1]


    @extent.setter
    def extent(self, extent):
        p1 = extent[:2]
        p2 = extent[2:]

        projection = self.provider.projection
        c1 = projection.rev_geocode(p1)
        c1 = zoom_to(c1, projection.zoom, self._zoom)

        c2 = projection.rev_geocode(p2)
        c2 = zoom_to(c2, projection.zoom, self._zoom)

        width = abs(c1[0] - c2[0]) * self.provider.tile_width
        height = abs(c1[1] - c2[1]) * self.provider.tile_height

        col = (c1[0] + c2[0]) / 2
        row = (c1[1] + c2[1]) / 2

        map_origin, map_offset = calculateMapCenter(self.provider, (col, row))
        self.origin = map_origin
        self.offset = map_offset
        self._size = int(width), int(height)


    @property
    def center(self):
        """
        Calculate map geographical center.

        It is a tuple of two values - longitude and latitude.

        Setting map geographical center affects map geographical extent.
        """
        w, h = self._size
        return self.geocode((w / 2, h / 2))


    @center.setter
    def center(self, center):
        projection = self.provider.projection
        c = projection.rev_geocode(center)
        c = zoom_to(c, projection.zoom, self._zoom)

        map_origin, map_offset = calculateMapCenter(self.provider, c)
        self.origin = map_origin
        self.offset = map_offset


    @property
    def zoom(self):
        """
        Map zoom value.

        Setting map value does *not* affect any other map properties like
        extent, center or image size.
        """
        return self._zoom


    @zoom.setter
    def zoom(self, zoom):
        self._change_center_zoom(self.center, zoom)


    @property
    def size(self):
        """
        Size of the image containing map.

        It is a sequence of two integer values - width and height of the
        image.

        Setting size of the image affects map geographical extent.
        """
        return self._size


    @size.setter
    def size(self, size):
        self._check_size(size)
        self._size = size

    def _check_size(self, size):
        """
        Check if `size` parameter has correct type.
        """
        valid = size is None \
            or len(size) == 2 \
            and all(isinstance(v, numbers.Integral) for v in tuple(size))
        if not valid:
            raise TypeError('Size should be sequence of two integer values')

    def _change_center_zoom(self, center, zoom):
        """
        Recalculate map to have new geographical center and update zoom
        value.

        :param center: Map geographical center.
        :param zoom: Map zoom value.
        """
        projection = self.provider.projection
        c = projection.rev_geocode(center)
        c = zoom_to(c, projection.zoom, zoom)

        map_origin, map_offset = calculateMapCenter(self.provider, c)
        self.origin = map_origin
        self.offset = map_offset
        self._zoom = zoom


    def _change_extent_and_zoom(self, extent):
        """
        Recalculate map to have new geographical extent and calculate map
        zoom.

        :param extent: Map geographical extent.
        """
        width, height = self._size

        p1 = extent[:2]
        p2 = extent[2:]

        map_origin, map_offset, zoom = calculateMapExtent(self.provider, width, height, p1, p2)
        self.origin = map_origin
        self.offset = map_offset
        self._zoom = zoom


    def __str__(self):
        return 'Map({provider}, {_size}, {origin}, {offset})'.format(**self.__dict__)


    def rev_geocode(self, location):
        """
        Reverse geocode geographical location.

        The method calculates location position (x, y) on map image.

        :param location: Geographical location (longitude, latitude).
        """
        ox, oy = self.offset
        projection = self.provider.projection
        coord = projection.rev_geocode(location)
        coord = zoom_to(coord, projection.zoom, self._zoom)

        # distance from the known coordinate offset
        x = ox + self.provider.tile_width * (coord[0] - self.origin[0])
        y = oy + self.provider.tile_height * (coord[1] - self.origin[1])

        # because of the center/corner business
        w, h = self.size
        return x + w / 2, y + h / 2


    def geocode(self, point):
        """
        Geocode map image point.

        The method calculates geographical location (longitude, latitude)
        of image point.

        :param point: Image map point (x, y).
        """
        max_zoom = MAX_ZOOM
        max_coord = zoom_to(self.origin, self._zoom, max_zoom)

        w, h = self._size

        # because of the center/corner business
        x, y = point[0] - w / 2, point[1] - h / 2

        # distance in tile widths from reference tile to point
        xtiles = (x - self.offset[0]) / self.provider.tile_width
        ytiles = (y - self.offset[1]) / self.provider.tile_height

        # distance in rows & columns at maximum zoom
        xDistance = xtiles * math.pow(2, (MAX_ZOOM - self._zoom))
        yDistance = ytiles * math.pow(2, (MAX_ZOOM - self._zoom))

        # new point coordinate reflecting that distance
        x = max_coord[0] + xDistance
        y = max_coord[1] + yDistance
        coord = zoom_to((x, y), max_zoom, self._zoom)

        projection = self.provider.projection
        location = projection.geocode(coord, self._zoom)

        return location


def render_map(map, tiles=None, downloader=None, **kw):
    """
    Download map tiles and render map image.

    If tiles are specified, then the provided tiles are used to render map.
    Otherwise, map tiles are downloaded from map provider.

    If `downloader` is null, then default map tiles downloader is used
    (:py:func:`geotiler.tile.io.fetch_tiles`).

    The function returns an image (instance of `PIL.Image` class).

    :param map: Map instance.
    :param tiles: Optional map tiles.
    :param downloader: Map tiles downloader.
    :param kw: Parameters passed to the downloader.
    """
    task = render_map_async(map, downloader=downloader, **kw)
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(task)

async def render_map_async(map, tiles=None, downloader=None, **kw):
    """
    Asyncio coroutine to download map tiles asynchronously and render map
    image.

    If tiles are specified, then the provided tiles are used to render map.
    Otherwise, map tiles are downloaded from map provider.

    If `downloader` is null, then default map tiles downloader is used
    (:py:func:`geotiler.tile.io.fetch_tiles`).

    The function returns an image (instance of `PIL.Image` class).

    :param map: Map instance.
    :param tiles: Optional map tiles.
    :param downloader: Map tiles downloader.
    :param kw: Parameters passed to the downloader.
    """
    if not tiles:
        tiles = fetch_tiles(map, downloader, **kw)
        tiles = (t async for t in tiles)
    return await render_image(map, tiles)

def fetch_tiles(map, downloader=None, **kw):
    """
    Create and fetch map tiles.

    Asynchronous generator of map tiles is returned.

    :param map: Map instance.
    :param downloader: Map tiles downloader.
    :param kw: Parameters passed to the downloader.
    """
    if downloader is None:
        downloader = _fetch_tiles

    tile_url = map.provider.tile_url

    coord, offset = _find_top_left_tile(map)
    coords = _tile_coords(map, coord, offset)
    offsets = _tile_offsets(map, offset)
    urls = (tile_url(c, map.zoom) for c in coords)
    tiles = (Tile(u, o, None, None) for u, o in zip(urls, offsets))
    return downloader(tiles, map.provider.limit, **kw)

def _tile_coords(map, coord, offset):
    """
    Create grid of coordinates of map tiles.

    :param map: Geotiler map instance.
    :param coord: Coordinate of top-left tile.
    :param offset: Map image offset of top-left tile.
    """
    w, h = map.size

    n = div_ceil(w - offset[0], map.provider.tile_width)
    m = div_ceil(h - offset[1], map.provider.tile_height)
    cols = range(coord[0], coord[0] + n)
    rows = range(coord[1], coord[1] + m)
    return itertools.product(cols, rows)


def _tile_offsets(map, offset):
    """
    Create grid of offsets of map tiles.

    :param map: Geotiler map instance.
    :param offset: Map image offset of top-left tile.
    """
    w, h = map.size
    cols = range(offset[0], w, map.provider.tile_width)
    rows = range(offset[1], h, map.provider.tile_height)
    return itertools.product(cols, rows)


def _find_top_left_tile(map):
    """
    Calculate the tile coordinate and its position relative to top-left
    offset of the map image.

    :param map: Map instance.
    """
    origin = map.origin
    w, h = map.size
    tw = map.provider.tile_width
    th = map.provider.tile_height

    px = map.offset[0] + w // 2
    py = map.offset[1] + h // 2

    dx = px // tw + (px % tw > 0)
    dy = py // th + (py % th > 0)

    px -= dx * tw
    py -= dy * th
    offset = px, py

    coord = origin[0] - dx, origin[1] - dy

    # if offset position not in ((-tw, -th), (0, 0)], then map image does
    # not contain the map properly
    assert -tw < offset[0] <= 0 and -th < offset[1] <= 0, offset

    return coord, offset


def calculateMapCenter(provider, tile_coord):
    """ Based on a provider and center coordinate, returns the coordinate
        of an initial tile and its point placement, relative to the map center.
    """
    # initial tile coordinate
    initTileCoord = math.floor(tile_coord[0]), math.floor(tile_coord[1])

    # initial tile position, assuming centered tile well in grid
    initX = (initTileCoord[0] - tile_coord[0]) * provider.tile_width
    initY = (initTileCoord[1] - tile_coord[1]) * provider.tile_height
    initPoint = round(initX), round(initY)

    return initTileCoord, initPoint

def calculateMapExtent(provider, width, height, *args):
    """ Based on a provider, width & height values, and a list of locations,
        returns the coordinate of an initial tile and its point placement,
        relative to the map center.
    """
    projection = provider.projection
    coordinates = [projection.rev_geocode(p) for p in args]

    TL = min(c[0] for c in coordinates), min(c[1] for c in coordinates)
    BR = max(c[0] for c in coordinates), max(c[1] for c in coordinates)

    # multiplication factor between horizontal span and map width
    hFactor = (BR[0] - TL[0]) / (width / provider.tile_width)

    # multiplication factor expressed as base-2 logarithm, for zoom difference
    hZoomDiff = math.log(hFactor) / math.log(2)

    # possible horizontal zoom to fit geographical extent in map width
    hPossibleZoom = projection.zoom - math.ceil(hZoomDiff)

    # multiplication factor between vertical span and map height
    vFactor = (BR[1] - TL[1]) / (height / provider.tile_height)

    # multiplication factor expressed as base-2 logarithm, for zoom difference
    vZoomDiff = math.log(vFactor) / math.log(2)

    # possible vertical zoom to fit geographical extent in map height
    vPossibleZoom = projection.zoom - math.ceil(vZoomDiff)

    # initial zoom to fit extent vertically and horizontally
    initZoom = min(hPossibleZoom, vPossibleZoom)

    ## additionally, make sure it's not outside the boundaries set by provider limits
    #initZoom = min(initZoom, provider.outerLimits()[1].zoom)
    #initZoom = max(initZoom, provider.outerLimits()[0].zoom)

    # coordinate of extent center
    centerColumn = (TL[0] + BR[0]) / 2
    centerRow = (TL[1] + BR[1]) / 2
    centerZoom = projection.zoom

    coord = zoom_to((centerColumn, centerRow), centerZoom, initZoom)

    origin, offset = calculateMapCenter(provider, coord)
    return origin, offset, initZoom


# vim: sw=4:et:ai



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
Create an object containing a GeoTiler object and convert the resulting
image to an OpenCV image in order to show it. Also contains functions to
plot points on the resulting map.

Zooming is possible with the mousewheel and `+` and `-` keys.

Requirements

- geotiler
- opencv
- numpy (also required by OpenCV)

Example provided by `https://github.com/matthijs876`.
"""

import geotiler
import cv2
import numpy as np


RED = (0, 0, 255)


class Map:
    """
    OpenCV compatible map image.

    After creating the object the map can be displayed with:

        kaart = Map((0.0, 51.47879))
        cv2.imShow('window name', Map)
        cv2.waitKey(10)

    :var markers: List of map markers.
    """
    def __init__(self, center, zoom=15, size=(1900, 1000)):
        """
        Create OpenCV compatible map image.

        :param center: Center of the map (longitude and latitude).
        :param zoom: Zoom level of the map.
        :param size: Resolution of the resulting image.
        """
        self.mm = geotiler.Map(center=center, zoom=zoom, size=size)
        self.markers = []
        self.update_map()


    def update_map(self):
        """
        Download new map tiles and redraw everyting on the map.
        """
        self.pil_image = geotiler.render_map(self.mm)
        self.draw_map()


    def draw_map(self):
        """
        Fetch map tiles and draw the map.

        The map markers are being plotted as well.

        Any additional drawing operations should be performed here.
        """
        data = np.array(self.pil_image)[:, :, :3]
        self.img = cv2.cvtColor(data, cv2.COLOR_RGB2BGR)
        self.plot_markers()


    def zoom_in(self):
        """
        Zoom in.

        Calling this function might fetch new tiles.
        """
        if self.mm.zoom < 19:
            self.mm.zoom += 1
        self.update_map()


    def zoom_out(self):
        """
        Zoom out.

        Calling this function might fetch new tiles.
        """
        if self.mm.zoom > 3:
            self.mm.zoom -= 1
        self.update_map()


    def plot_markers(self):
        """
        Draws all markers on the map.

        :param markers: Collection of map markers.
        """
        positions = (self.mm.rev_geocode(p) for p in self.markers)
        positions = ((int(lon), int(lat)) for lon, lat in positions)
        for pos in positions:
            cv2.circle(self.img, center=pos, radius=5, color=RED, thickness=-1)


    def mouse_callback(self, event, x, y, flag=0, param=None):
        """
        Mouse events callback.

        Events handled

        left-click
            add map marker
        right-click
            remove map marker
        scroll
            zoom in or out at the location of the mouse pointer
        """
        if event == cv2.EVENT_MOUSEWHEEL:
            if flag > 0:  # Scroll up
                self.mm.center = self.mm.geocode((x, y))
                self.zoom_in()
            elif flag < 0:    # Scroll down
                self.mm.center = self.mm.geocode((x, y))
                self.zoom_out()
        elif event == cv2.EVENT_LBUTTONUP:
            self.markers.append(self.mm.geocode((x, y)))
            self.draw_map()
        elif event == cv2.EVENT_RBUTTONUP:
            if self.markers:
                del self.markers[-1]
                self.draw_map()
            else:
                print('Nothing to delete')



# Create the map object (`kaart` is map in Dutch)
kaart = Map((0.0, 51.47879))

# Create a window and adjust its size automatically
cv2.namedWindow('window', cv2.WINDOW_AUTOSIZE)

# Create the mousec allback in the window
cv2.setMouseCallback('window', kaart.mouse_callback)

while 1:
    # Show the image in the window
    cv2.imshow('window', kaart.img)

    # OpenCV doesn't show anything until the `waitKey` function is called
    key = cv2.waitKey(20)

    # when Esc is pressed: close all opencv windows and break
    if key == 27:
        cv2.destroyAllWindows()
        break
    elif key == 43:  # "+" key
        kaart.zoom_in()
    elif key == 45:   # "-" key
        kaart.zoom_out()

# vim: sw=4:et:ai



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
Render map image using map tile data.
"""

import io
import functools
import logging

import PIL.Image  # type: ignore
import PIL.ImageDraw  # type: ignore

logger = logging.getLogger(__name__)

async def render_image(map, tiles):
    """
    Redner map image using map tile data.

    Each item in tile data collection is tile image data, which can be
    interpreted with `PIL` library (via `PIL.Image.open` call, i.e. PNG
    file data or JPEG file data). The item can also be `None` if tile data
    could not be downloaded, i.e. due to network error.

    The map tiles are rendered into single map image. Error tile image is
    rendered if data for a tile does not exist.

    The PIL image object is returned.

    :param map: Map object.
    :param tiles: Asynchronous generator of map tiles.
    """
    if __debug__:
        logger.debug('combining tiles')

    provider = map.provider

    # PIL requires image size to be a tuple
    image = PIL.Image.new('RGBA', tuple(map.size))
    error = _error_image(provider.tile_width, provider.tile_height)

    async for tile in tiles:
        img = _tile_image(tile.img) if tile.img else error
        image.paste(img, tile.offset)

    return image

@functools.lru_cache(maxsize=4)
def _error_image(width, height):
    """
    Create error tile image.

    The error tile image is PIL image object showing message that a map
    tile could not be downloaded.

    :param width: Width of tile image.
    :param height: Height of tile image.
    """
    img = PIL.Image.new('RGBA', (width, height))
    draw = PIL.ImageDraw.Draw(img)
    msg = 'Error downloading map tile.'
    tw, th = draw.textsize(msg)
    draw.text(((width - tw) // 2, (height - th) // 2), msg, 'red')
    return img

def _tile_image(data):
    """
    Convert image data like PNG file data or JPEG file data into
    `PIL.Image` object.

    :param data: Tile data, i.e. PNG file data.
    """
    f = io.BytesIO(data)
    return PIL.Image.open(f).convert('RGBA')


# vim: sw=4:et:ai



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
Load and create map providers.
"""

import configparser
import glob
import itertools
import json
import logging
import os.path
import re
import typing as tp

from .geo import WebMercator
from .errors import GeoTilerError

logger = logging.getLogger(__name__)

DEFAULT_PROVIDER = 'osm'
RE_URL_OBFUSCATE = re.compile('(?<=apikey=)[a-z0-9-]+|(?<=api-key=)[a-z0-9-]+', re.I)

# the attributes inspired by poor-maps project tile source definition
# https://github.com/otsaloma/poor-maps/tree/master/tilesources
ATTRIBUTES = 'id', 'name', 'attribution', 'url', 'subdomains', 'extension', \
    'limit', 'api-key-ref', 'tile-width', 'tile-height'

class MapProvider:
    def __init__(self, data: tp.Dict, api_key: tp.Optional[str]=None) -> None:
        self.id = None
        self.name = None
        self.attribution = None
        self.url = None
        self.subdomains: tp.Tuple[str, ...] = tuple()
        self.extension = 'png'
        self.limit = 1
        self.api_key_ref = None
        self.api_key = api_key
        self.tile_width = 256
        self.tile_height = 256

        # change a-b-c to a_b_c to allow python attribute access
        norm = lambda n: n.replace('-', '_')
        attrs = ((norm(n), data[n]) for n in ATTRIBUTES if n in data)
        self.__dict__.update(attrs)

        self.projection = WebMercator(0)
        if self.subdomains:
            self.subdomain_cycler = itertools.cycle(self.subdomains)
        else:
            self.subdomain_cycler = itertools.cycle(('', ))

    def tile_url(self, tile_coord, zoom):
        params = {
            'subdomain': next(self.subdomain_cycler),
            'x': tile_coord[0],
            'y': tile_coord[1],
            'z': zoom,
            'ext': self.extension,
            'api_key': self.api_key,
        }
        url = self.url.format(**params)
        if __debug__:
            logger.debug('tile url: {}'.format(obfuscate(url)))
        return url

    def __str__(self):
        return self.name

def providers():
    """
    Get sorted list of all map providers identificators.
    """
    path = os.path.join(base_dir(), '*.json')
    if __debug__:
        logger.debug('list map providers from {}'.format(path))
    pid = lambda fn: os.path.splitext(os.path.basename(fn))[0]
    return sorted(pid(fn) for fn in glob.iglob(path))

def find_provider(id):
    """
    Load map provider data from JSON file and create map provider.

    :param id: Map provider identificator.
    """
    data = read_provider_data(id)

    api_key_ref = data.get('api-key-ref')
    api_key = None
    if api_key_ref:
        cp = read_config()

        # no api key for the api key reference, then raise fatal error; no
        # api key means no access
        if not cp.has_option('api-key', api_key_ref):
            raise GeoTilerError('No API key for for reference "{}"'.format(api_key_ref))

        api_key = cp.get('api-key', api_key_ref)

    if __debug__:
        logger.debug('map provider "{}" api key reference: {}'.format(
            id, api_key_ref
        ))

    provider = MapProvider(data, api_key=api_key)
    return provider

def base_dir() -> str:
    """
    Get map provider data base directory.
    """
    mod = __import__('geotiler')
    return os.path.join(mod.__path__[0], 'source')

def read_config() -> configparser.ConfigParser:
    """
    Read GeoTiler configuration file.
    """
    p = os.getenv('HOME', '')
    fn = os.path.join(p, '.config/geotiler/geotiler.ini')
    if not os.path.exists(fn):
        raise GeoTilerError('Configuration file {} not found'.format(fn))

    cp = configparser.ConfigParser()
    cp.read(fn)
    return cp

def read_provider_data(id: str) -> tp.Dict:
    """
    Read map provider data.

    :param id: Map provider identificator.
    """
    fn = os.path.join(base_dir(), id + '.json')
    if __debug__:
        logger.debug('loading map provider "{}" from {}'.format(id, fn))

    with open(fn, encoding='utf8') as f:
        data = json.load(f)

    return data

def obfuscate(url: str) -> str:
    """
    Replace API key in a tile URL with "<apikey>" string.

    :param url: Tile URL to obfuscate.
    """
    return RE_URL_OBFUSCATE.sub('<apikey>', url)

# vim:et sts=4 sw=4:



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

import pkg_resources

from .map import Map, render_map, render_map_async, fetch_tiles
from .provider import find_provider, providers

__version__ = pkg_resources.get_distribution('geotiler').version

# vim: sw=4:et:ai



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
Functions and coroutines to download map tiles.
"""

import aiohttp
import asyncio
import logging
import pkg_resources
from functools import partial

logger = logging.getLogger(__name__)

HEADERS = {
    'User-Agent': 'GeoTiler/{}'.format(pkg_resources.get_distribution('geotiler').version),
}

# client session params
PARAMS = {
    'headers': HEADERS,
    'trust_env': True,
    'raise_for_status': True,
}

FMT_DOWNLOAD_LOG = 'Cannot download a tile due to error: {}'.format
FMT_DOWNLOAD_ERROR = 'Unable to download {} (error: {})'.format

async def fetch_tile(session, tile):
    """
    Fetch map tile.

    :param tile: Map tile.
    """
    try:
        async with session.get(tile.url) as response:
            data = await response.read()
    except aiohttp.ClientError as ex:
        error = ValueError(FMT_DOWNLOAD_ERROR(tile.url, ex))
        tile = tile._replace(img=None, error=error)
    else:
        tile = tile._replace(img=data, error=None)

    return tile

async def fetch_tiles(tiles, num_workers):
    """
    Download map tiles.

    Asynchronous generator of map tiles is returned.

    A collection of tiles is returned. Each successfully downloaded tile
    has `Tile.img` attribute set. If there was an error while downloading
    a tile, then `Tile.img` is set to null and `Tile.error` to a value error.

    :param tiles: Collection of tiles.
    :param num_workers: Number of workers used to connect to a map provider
        service.
    """
    if __debug__:
        logger.debug('fetching tiles...')

    # respect connection limits by defining custom connector
    connector = aiohttp.TCPConnector(limit_per_host=num_workers)

    # use `trust_env` to get proxy configuration via env variables
    async with aiohttp.ClientSession(connector=connector, **PARAMS) as session:
        f = partial(fetch_tile, session)
        tasks = [f(t) for t in tiles]
        for task in asyncio.as_completed(tasks):
            tile = await task  # no exception expected at this stage

            if tile.error:
                logger.warning(FMT_DOWNLOAD_LOG(tile.error))

            yield tile

    if __debug__:
        logger.debug('fetching tiles done')

# vim: sw=4:et:ai
